'''Useful functions

    date_from_relative - convert relative period like '1m' to past date

    convert_to_edge - translate list of vcs.ModifiedFile to vsc.EdgeFile
'''
from __future__ import annotations

from datetime import datetime
import re

from dateutil import relativedelta

from smyg import vcs


def date_from_relative(past: str) -> datetime:
    '''convert relative period like '1m' to past date

    date_from_relative('1m') => TODAY - 1 month
    date_from_relative('21d') => TODAY - 28 days
    etc
    '''
    if not re.match(r'^\d+[ymd]$', past):
        raise ValueError('Relative argument does not match pattern')
    period_types = {'y': 'years', 'm': 'months', 'd': 'days'}
    period = {period_types[past[-1]]: int(past[:-1])}
    return datetime.now() - relativedelta.relativedelta(**period)


def convert_to_edge(modified_files: list[vcs.ModifiedFile]
                    ) -> list[vcs.EdgeFile]:
    '''convert modified files to edge files'''
    edge_files = []
    for modified_file in modified_files:
        for edge_file in edge_files:
            if edge_file.path == modified_file.prev_path:
                edge_file.update(modified_file)
                break
        else:
            edge_files.append(vcs.EdgeFile(modified_file))
    return edge_files



'''Prometheus client'''
from __future__ import annotations

import os

import prometheus_client


PUSHGATEWAY_URL = os.getenv('PUSHGATEWAY_URL', 'http://localhost:9091')
PUSHGATEWAY_USERNAME = os.getenv('PUSHGATEWAY_USERNAME')
PUSHGATEWAY_PASSWORD = os.getenv('PUSHGATEWAY_PASSWORD')
PROJECT_NAME = os.getenv('PROJECT_NAME')

registry = prometheus_client.CollectorRegistry()

COMMIT_LABELS = ['author_name', 'author_email', 'project_name']

COMMITS = prometheus_client.Counter(
        'commits',
        'Number of commits',
        COMMIT_LABELS,
        registry=registry)
COMMIT_ADDED_LINES = prometheus_client.Counter(
        'commit_added_lines',
        'Number of added lines in commit',
        COMMIT_LABELS,
        registry=registry)
COMMIT_DELETED_LINES = prometheus_client.Counter(
        'commit_deleted_lines',
        'Number of added deleted in commit',
        COMMIT_LABELS,
        registry=registry)
COMMIT_CHANGED_FILES = prometheus_client.Counter(
        'commit_changed_files',
        'Number of changed files in commit',
        COMMIT_LABELS,
        registry=registry)


class PushGatewayError(Exception):
    '''Default exception for prom module'''


def push_commits(values: list | dict):
    '''Push commit values to pushgateway'''
    if not PUSHGATEWAY_URL:
        raise PushGatewayError(
                '\nAn error occurred during the push metrics:\n'
                'PUSHGATEWAY_URL env variable not set')
    if not isinstance(values, list):
        values = [values]
    for value in values:
        labels = (value.get('author_name'),
                  value.get('author_email'),
                  PROJECT_NAME)
        COMMITS.labels(*labels).inc()
        COMMIT_ADDED_LINES.labels(*labels).inc(value.get('added'))
        COMMIT_DELETED_LINES.labels(*labels).inc(value.get('deleted'))
        COMMIT_CHANGED_FILES.labels(*labels).inc(value.get('changed_files'))
    prometheus_client.push_to_gateway(
            PUSHGATEWAY_URL,
            job='showmeyourgit',
            registry=registry)


def push_codechanges(value: dict,
                     metric_name: str = 'codechanges',
                     suffix: str = ''):
    '''Push codechurn values to pushgateway

    Args:
        value - metrics value
        metric_name - metric name: codechanges, codechurn. Uses in metrics
            with equal value format
        suffix - metrics name suffix with relative interval,
            e.g. '5d', '3m'
    '''
    if suffix:
        suffix = '_' + suffix
    added_lines = prometheus_client.Gauge(
            f'{metric_name}_added_lines{suffix}',
            'Number of added lines for period',
            ['project_name'],
            registry=registry)
    added_lines.labels(PROJECT_NAME).set(value.get('added'))
    deleted_lines = prometheus_client.Gauge(
            f'{metric_name}_deleted_lines{suffix}',
            'Number of deleted lines from those that were added for period',
            ['project_name'],
            registry=registry)
    deleted_lines.labels(PROJECT_NAME).set(value.get('deleted'))
    ratio = prometheus_client.Gauge(
            f'{metric_name}_ratio{suffix}',
            'Relative churn value for all files',
            ['project_name'],
            registry=registry)
    ratio.labels(PROJECT_NAME).set(value.get('ratio'))
    prometheus_client.push_to_gateway(
            PUSHGATEWAY_URL,
            job=f'showmeyourgit-{PROJECT_NAME}-{metric_name}{suffix}',
            registry=registry)


def push_gateway_handler(url, method, timeout, headers, data):
    return prometheus_client.exposition.basic_auth_handler(
            url,
            method,
            timeout,
            headers,
            data,
            PUSHGATEWAY_USERNAME,
            PUSHGATEWAY_PASSWORD)



'''Internal entities of the git vcs, like commits, modified files, tags
and support classes.
'''

from __future__ import annotations

import dataclasses
import datetime
import enum
import hashlib


def ratio(added: int, deleted: int) -> float:
    '''Calculate deleted and added lines ratio'''
    if added == 0:
        return 0.0
    return round(deleted / added * 100, 2)


class ChangeType(enum.Enum):
    '''Type of modified file change'''
    UNDEFINED = 0
    ADDED = 1
    DELETED = 2
    MODIFIED = 3
    RENAMED = 4


class SourceLine:
    '''Line instance of source file with number and text value

    Attributes:
        number: number of line
        value: line text
    '''

    def __init__(self, number: int, value: str) -> None:
        self.number = number
        self.value = value
        self.digest = hashlib.md5(value.encode()).hexdigest()

    def __eq__(self, other) -> bool:
        if self.digest == other.digest:
            return True
        return False

    def __repr__(self) -> str:
        return f'{self.number:5d}: {self.value}'


def parsed_diff_to_source_lines(parsed_diff: list[tuple]) -> list[SourceLine]:
    '''Convert added or deleted lines to list of SourceLine instances

    Args:
        parsed_diff: added or deleted sequence of pairs:
            [(LINE_NUM, LINE_TEXT), ...]
    Returns:
        list of SourceLine instances.
    '''
    lines = []
    for diff in parsed_diff:
        lines.append(SourceLine(number=diff[0], value=diff[1]))
    return lines


class ModifiedFile:
    '''Modified file info with added or deleted lines

    Attributes:
        added: number of added lines
        deleted: number of deleted lines
        changed_lines_total: added and deleted lines sum
        path: actual file path (after rename or delete)
        prev_path: previous file path (before rename or delete)
        path_changed: file path was changed (renamed)
        change_type: file was added, deleted, renamed or just modified
    '''

    def __init__(self,
                 old_path: str | None,
                 new_path: str | None,
                 added_lines: list[tuple] = None,
                 deleted_lines: list[tuple] = None):
        '''Object initalizer

        Args:
            old_path - source file path before changes
            new_path - source file path after changes
            added_lines - added lines to file: [(LINE_NUM, LINE_TEXT), ...]
            deleted_lines - deleted lines from file:
                [(LINE_NUM, LINE_TEXT), ...]
        '''
        self.old_path = old_path
        self.new_path = new_path
        self.added_lines = parsed_diff_to_source_lines(added_lines or [])
        self.deleted_lines = parsed_diff_to_source_lines(deleted_lines or [])

    @property
    def added(self) -> int:
        return len(self.added_lines)

    @property
    def deleted(self) -> int:
        return len(self.deleted_lines)

    @property
    def changed_lines_total(self) -> int:
        return self.added + self.deleted

    @property
    def path(self) -> str:
        if self.change_type == ChangeType.DELETED:
            return self.old_path
        return self.new_path

    @property
    def prev_path(self) -> str:
        if self.path_changed:
            return self.old_path
        return self.path

    @property
    def path_changed(self) -> bool:
        if self.old_path and self.new_path and self.old_path != self.new_path:
            return True
        return False

    @property
    def change_type(self) -> ChangeType:
        if self.old_path is None and self.new_path:
            return ChangeType.ADDED
        if self.old_path and self.new_path is None:
            return ChangeType.DELETED
        if self.old_path and self.new_path and self.changed_lines_total > 0:
            return ChangeType.MODIFIED
        if self.old_path and self.new_path and self.changed_lines_total == 0:
            return ChangeType.RENAMED
        return ChangeType.UNDEFINED

    def __repr__(self) -> str:
        if self.path_changed:
            return f'{self.old_path}->{self.new_path} {self.change_type}'
        return f'{self.path} {self.change_type}'


class EdgeFile:
    '''Edge file state with actual numbers of added lines

    Every modified file instance change state: lines added, lines deleted,
    file path changed.
    Every modification in lines shifts line numbers

    Attributes:
        added: total number of added lines
        deleted: total number of deleted lines from added lines
        ratio: the ratio between deleted and added linese
        churn: number of deleted lines from added lines
        churn_ratio: the ratio between churn and added linese (code churn)
        line_value: get line text by line number
    '''

    def __init__(self, file: ModifiedFile):
        # --- total added lines
        self.added_lines = []
        # --- deleted lines from added lines
        self.deleted_lines = []
        # --- churn lines
        self.churn_lines = []
        # ---
        self.update(file)

    def update(self, file: ModifiedFile):
        '''Update current path and state of lines'''
        self.path = file.path

        # --- process deleted lines
        for line in file.deleted_lines:
            self.deleted_lines.append(line)
            if line in self.added_lines:
                self.churn_lines.append(line)

        # --- process added lines
        for line in file.added_lines:
            self.added_lines.append(line)

    @property
    def added(self) -> int:
        return len(self.added_lines)

    @property
    def deleted(self) -> int:
        return len(self.deleted_lines)

    @property
    def churn(self) -> int:
        return len(self.churn_lines)

    @property
    def ratio(self) -> float:
        return ratio(added=self.added, deleted=self.deleted)

    @property
    def churn_ratio(self) -> float:
        return ratio(added=self.added, deleted=self.churn)

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path


# pylint: disable=too-many-instance-attributes
@dataclasses.dataclass
class Commit:
    '''A Commit object has all the information of a Git commit'''

    sha: str
    msg: str
    author_name: str
    author_email: str
    author_date: datetime.datetime
    committer_name: str
    committer_email: str
    committer_date: datetime.datetime
    project_name: str
    added: int
    deleted: int
    changed_files: int
    branches: list

    def as_dict(self) -> dict:
        '''Represent instance as dict'''
        return dataclasses.asdict(self)

    def as_seriable_dict(self) -> dict:
        '''Represent instance as seriable dict'''
        seriable = dict(self.as_dict())
        seriable.update(
                {'author_date': str(self.author_date),
                 'committer_date': str(self.committer_date),
                 'branches': list(self.branches),
                 })
        return seriable

    def __repr__(self):
        return f'{self.sha} {self.msg}'



'''Code changes metric module

   Metrics values:
      * count of added lines
      * count of deleted lines
      * relative changes value for each file
      * relative changes value for all files
'''

from typing import Mapping, List

from smyg import vcs
from smyg import utils


class CodeChanges:
    '''This metric measures the code changes of files and relative values

    Attributes:
        calculate: calculate metric values
    '''

    def __init__(self,
                 modified_files: List[vcs.ModifiedFile] = None):
        '''Object initalizer

        Attributes:
            commits - chronological sequence of commits (modified files groups)
        '''
        self.modified_files = modified_files or []

    def calculate(self, detail: bool = False) -> Mapping:
        '''Calculate code changes metric

        Args:
            detail: detail code changes per each file

        Returns:
            A dict with calculated values.
            For example:

            {'added': 1000,
             'deleted': 200,
             'ratio': '20',
             'files': [('/file/path': 10, 1, 10')]}

            Returned keys:
            added: count of added lines
            deleted: count of deleted lines from those that were added
            ratio: relative value for all files (in percent)
            files: detailed info per each file
        '''

        self.edge_files = utils.convert_to_edge(self.modified_files)

        # --- calculate lines in edge files
        added = deleted = 0
        detail_files = []
        for edge_file in self.edge_files:
            added += edge_file.added
            deleted += edge_file.deleted
            detail_files.append((edge_file.path,
                                 edge_file.added,
                                 edge_file.deleted,
                                 edge_file.ratio))

        values = {'added': added,
                  'deleted': deleted,
                  'ratio': vcs.ratio(added=added, deleted=deleted)}

        if detail:
            values['files'] = sorted(detail_files,
                                     key=lambda x: x[3],
                                     reverse=True)

        return values



'''Covert metrics or instance dictionaries to specified format'''

from __future__ import annotations

import os
import json
import jinja2


class FormatterError(Exception):
    '''Default formatter exception'''


def render(value: dict, output: str, template_name: str = None) -> str:
    '''Convert metrics to output format'''
    if output == 'text':
        path = os.path.join(os.path.dirname(__file__), 'templates')
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(
            searchpath=path))
        template = env.get_template(f'{template_name}.j2')
        return template.render({'ctx': value})

    if output == 'json':
        return json.dumps(value)

    raise FormatterError(f'Unknown format {output}')



'''command line interface'''
from __future__ import annotations

import datetime as dt
import os
import typer

from smyg import __app_name__, __version__
from smyg import binding
from smyg import formatter
from smyg import prom
from smyg import utils
from smyg.metrics import code_churn as cc
from smyg.metrics import code_changes as cch

app = typer.Typer()


@app.command()
def branch_commits(
        sha: str =
        typer.Argument(
            None,
            help='beginning commit SHA'),
        repo_path: str =
        typer.Option(
            '.',
            '--repo-path',
            help='path to git repo'),
        branch: str =
        typer.Option(
            None,
            help='show commits for specified branch'
            ),
        output: str =
        typer.Option(
            'text',
            '--output',
            '-o',
            help='output format [text/json]'),
        push_metrics: bool =
        typer.Option(
            False,
            help='push metrics to gateway'),
        ):
    '''Show changes for bundle of commits

    Assumed that will be used in CI on push event job.
    If we have previous SHA then using just commits with --from-commit.
    If we don't have SHA or '000...' as SHA value then we use only commits
    in single branch
    '''
    path = os.path.abspath(repo_path)
    # --- In Gitlab CI beginning sha contains '000.' (see CI_COMMIT_BEFORE_SHA)
    if sha == '0000000000000000000000000000000000000000':
        sha = None
    # --- get commits bundle
    try:
        commits = binding.branch_commits(path, sha=sha, branch=branch)
    except binding.BindingError as err:
        raise typer.Exit(err)

    typer.echo(
            formatter.render(
                [commit.as_seriable_dict() for commit in commits],
                output,
                'commits'
                ))

    if push_metrics:
        try:
            prom.push_commits([commit.as_dict() for commit in commits])
        except prom.PushGatewayError as err:
            raise typer.Exit(err)


@app.command()
def commit(
        sha: str =
        typer.Argument(
            None,
            help='hash of the commit'),
        repo_path: str =
        typer.Option(
            '.',
            '--repo-path',
            help='path to git repo'),
        branch: str =
        typer.Option(
            None,
            help='show commit info in last commit in branch'
            ),
        output: str =
        typer.Option(
            'text',
            '--output',
            '-o',
            help='output format [text/json]'),
        push_metrics: bool =
        typer.Option(
            False,
            help='push metrics to gateway'),
        ):
    '''Show commit info'''
    path = os.path.abspath(repo_path)
    try:
        vcs_commit = binding.find_commit(path, sha=sha, branch=branch)
    except binding.BindingError as err:
        raise typer.Exit(err)
    typer.echo(
            formatter.render(vcs_commit.as_seriable_dict(), output, 'commit'))
    if push_metrics:
        try:
            prom.push_commits(vcs_commit.as_seriable_dict())
        except prom.PushGatewayError as err:
            raise typer.Exit(err)


@app.command()
def codechanges(
        repo_path: str =
        typer.Option(
            '.',
            '--repo-path',
            help='path to git repo'),
        from_date: dt.datetime =
        typer.Option(
            None,
            help='only commits after this date will be analyzed',
            rich_help_panel='FROM'),
        from_commit: str =
        typer.Option(
            None,
            '--from-commit',
            help='only commits after this commit hash '
            'will be analyzed',
            rich_help_panel='FROM'),
        from_tag: str =
        typer.Option(
            None,
            '--from-tag',
            help='only commits after this commit tag will be analyzed',
            rich_help_panel='FROM'),
        to_date: dt.datetime =
        typer.Option(
            None,
            help='only commits up to this date will be analyzed',
            rich_help_panel='TO'),
        to_commit: str =
        typer.Option(
            None,
            '--to-commit',
            help='only commits up to this commit hash will be analyzed',
            rich_help_panel='TO'),
        to_tag: str =
        typer.Option(
            None,
            '--to-tag',
            help='only commits up to this commit tag will be analyzed',
            rich_help_panel='TO'),
        branch: str =
        typer.Option(
            None,
            help='only analyses commits that belong to this branch'
            ),
        detail: bool =
        typer.Option(
            False,
            help='show detail code churn per each file'),
        output: str =
        typer.Option(
            'text',
            '--output',
            '-o',
            help='output format [text/json]'),
        for_past: str =
        typer.Option(
            None,
            help='metric for relative period, e.g. 5d, 3m, 1y'),
        push_metrics: bool =
        typer.Option(
            False,
            help='push metrics to prometheus gateway'),
            ):
    '''This metric measures the code changes of each files.'''

    path = os.path.abspath(repo_path)
    if for_past:
        try:
            from_date = utils.date_from_relative(for_past)
        except ValueError as err:
            raise typer.Exit(err)
    try:
        modified_files = binding.modified_files(path,
                                                since=from_date,
                                                from_commit=from_commit,
                                                from_tag=from_tag,
                                                to=to_date,
                                                to_commit=to_commit,
                                                to_tag=to_tag,
                                                only_in_branch=branch)
    except binding.BindingError as err:
        raise typer.Exit(err)
    metrics = cch.CodeChanges(modified_files).calculate(detail=detail)
    typer.echo(formatter.render(metrics, output, 'code_changes'))
    if push_metrics:
        try:
            prom.push_codechanges(metrics,
                                  suffix=for_past)
        except prom.PushGatewayError as err:
            raise typer.Exit(err)


@app.command()
def codechurn(
        repo_path: str =
        typer.Option(
            '.',
            '--repo-path',
            help='path to git repo'),
        from_date: dt.datetime =
        typer.Option(
            None,
            help='only commits after this date will be analyzed',
            rich_help_panel='FROM'),
        from_commit: str =
        typer.Option(
            None,
            '--from-commit',
            help='only commits after this commit hash '
            'will be analyzed',
            rich_help_panel='FROM'),
        from_tag: str =
        typer.Option(
            None,
            '--from-tag',
            help='only commits after this commit tag will be analyzed',
            rich_help_panel='FROM'),
        to_date: dt.datetime =
        typer.Option(
            None,
            help='only commits up to this date will be analyzed',
            rich_help_panel='TO'),
        to_commit: str =
        typer.Option(
            None,
            '--to-commit',
            help='only commits up to this commit hash will be analyzed',
            rich_help_panel='TO'),
        to_tag: str =
        typer.Option(
            None,
            '--to-tag',
            help='only commits up to this commit tag will be analyzed',
            rich_help_panel='TO'),
        branch: str =
        typer.Option(
            None,
            help='only analyses commits that belong to this branch'
            ),
        detail: bool =
        typer.Option(
            False,
            help='show detail code churn per each file'),
        output: str =
        typer.Option(
            'text',
            '--output',
            '-o',
            help='output format [text/json]'),
        for_past: str =
        typer.Option(
            '',
            help='metric for relative period, e.g. 5d, 3m, 1y'),
        push_metrics: bool =
        typer.Option(
            False,
            help='push metrics to prometheus gateway'),
        ):
    '''This metric measures the code churns of each file.'''

    path = os.path.abspath(repo_path)
    if for_past:
        try:
            from_date = utils.date_from_relative(for_past)
        except ValueError as err:
            raise typer.Exit(err)
    try:
        modified_files = binding.modified_files(path,
                                                since=from_date,
                                                from_commit=from_commit,
                                                from_tag=from_tag,
                                                to=to_date,
                                                to_commit=to_commit,
                                                to_tag=to_tag,
                                                only_in_branch=branch)
    except binding.BindingError as err:
        raise typer.Exit(err)
    code_churn = cc.CodeChurn(modified_files)
    metrics = code_churn.calculate(detail=detail)
    typer.echo(formatter.render(metrics, output, 'code_changes'))
    if push_metrics:
        try:
            prom.push_codechanges(metrics,
                                  metric_name='codechurn',
                                  suffix=for_past)
        except prom.PushGatewayError as err:
            raise typer.Exit(err)


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f'{__app_name__} v{__version__}')
        raise typer.Exit()


@app.callback()
def main(
    version=typer.Option(
        None,
        '--version',
        '-v',
        help="show the application's version and exit",
        callback=_version_callback,
        is_eager=True,
    )
) -> None:
    # pylint: disable=unused-argument,missing-function-docstring
    return



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smyg', 'smyg.metrics']

package_data = \
{'': ['*'], 'smyg': ['templates/*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0',
 'prometheus-client>=0.16,<0.17',
 'pydriller>=2.3,<3.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'rich>=13.1.0,<14.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['smyg = smyg.cli:app']}

setup_kwargs = {
    'name': 'smyg',
    'version': '0.5.0',
    'description': 'git metric calculation',
    'long_description': '# ShowMeYourGit\n\n**showmeyourgit** (or smyg) is a project for calculating git metrics: added lines, deleted lines, commits count, changed files, code changes ratio, code churn ratio.\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install smyg.\n\n```bash\npip install --user smyg\n```\n\n## Usage\n\nClone [example project](https://github.com/nmix/example-project)\n```bash\ngit clone https://github.com/nmix/example-project.git\ncd example-project\n```\n\n### commit\n\n Show commit info\n\n```bash\n# --- show command help\n$ smyg commit --help\n\n# --- show last commit info for current branch\n$ smyg commit\ncdcb7c6ac0b92b1e16cfdd6f1e6abf0ed8f73e48 | update README\n2023-02-20 16:42:41+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:       26\nDeleted lines:      0\nChanged files:      1\n\n# --- show last commit info in json format\n$ smyg commit --output json | jq\n{\n  "hash": "cdcb7c6ac0b92b1e16cfdd6f1e6abf0ed8f73e48",\n  "msg": "update README",\n  "author_name": "Nikolay Mikhaylichenko",\n  "author_email": "nn.mikh@yandex.ru",\n  "author_date": "2023-02-20 16:42:41+03:00",\n  "committer_name": "Nikolay Mikhaylichenko",\n  "committer_email": "nn.mikh@yandex.ru",\n  "committer_date": "2023-02-20 16:42:41+03:00",\n  "project_name": "example-project",\n  "added": 26,\n  "deleted": 0,\n  "changed_files": 1,\n  "branches": [\n    "main"\n  ]\n}\n\n# --- show info about arbitary commit\n$ smyg commit 07cb19ceba138f1d7e4a1d48f024e520f3657a1d\n07cb19ceba138f1d7e4a1d48f024e520f3657a1d | commit 2-1\n2023-02-20 16:26:48+03:00 | Nikolay Mikhaylichenko\n[\'branch-2\']\nAdded lines:        6\nDeleted lines:      0\nChanged files:      1\n```\n\n### branch-commits\n\nShow changes for bundle of commits in branch.\n\nIf we know last commit of previous bundle then we take all commits after that. In Gitlab CI previous commit SHA may be in `CI_COMMIT_BEFORE_SHA` environment variable (see [Predefined variables reference](https://docs.gitlab.com/ee/ci/variables/predefined_variables.html)).\n\nIf we dont have previous commit then we take only commits for *single* branch. *Signle* branch means that commit belongs only one branch. **Here it is mandatory to have a complete local copy of the repository with all branches.**\n\n```bash\n# --- show command help\n$ smyg branch-commits --help\n\n# --- if you just clone the example repository, the entire commit list will be displayed (from Initial commit)\n#     this is because only one branch is locally received and all commits belongs to it\n$ smyg branch-commits\n\ncdcb7c6ac0b92b1e16cfdd6f1e6abf0ed8f73e48 | update README\n2023-02-20 16:42:41+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:       26\nDeleted lines:      0\nChanged files:      1\n\n7febdeea7b4f82d15f7fa0765c12f10928af7a1a | lorem ipsum in main\n2023-02-20 16:33:58+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:        6\nDeleted lines:      0\nChanged files:      1\n\n# ... skip remaining commits\n\n# --- checkout branch-3 and repeat the command\n#     we get only 3 commits for the branch\n$ git checkout branch-3\nSwitched to a new branch \'branch-3\'\n$ smyg branch-commits\n\n414d13a92998f9de86dd60d41f05f092e64350f3 | commit 3-3\n2023-02-20 16:32:42+03:00 | Nikolay Mikhaylichenko\n[\'branch-3\']\nAdded lines:        2\nDeleted lines:      0\nChanged files:      1\n\n776206b033b8a149c4ee491ec9bee23273f98afc | commit 3-2\n2023-02-20 16:32:13+03:00 | Nikolay Mikhaylichenko\n[\'branch-3\']\nAdded lines:        2\nDeleted lines:      0\nChanged files:      1\n\naa4f97ed8366b8d72a0be2c1c114aa503b531190 | commit 3-1\n2023-02-20 16:31:54+03:00 | Nikolay Mikhaylichenko\n[\'branch-3\']\nAdded lines:        4\nDeleted lines:      0\nChanged files:      1\n\n# --- branch commits with previous SHA\n$ git checkout main\nsmyg branch-commits 5317cd6080033ae5a9ce3166095745e862a52c9d\n\ncdcb7c6ac0b92b1e16cfdd6f1e6abf0ed8f73e48 | update README\n2023-02-20 16:42:41+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:       26\nDeleted lines:      0\nChanged files:      1\n\n7febdeea7b4f82d15f7fa0765c12f10928af7a1a | lorem ipsum in main\n2023-02-20 16:33:58+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:        6\nDeleted lines:      0\nChanged files:      1\n\nbbead3ba03de57909ee622fac4b973de180f5e5f | Branch 2 (#2)\n\ncommit 2-1  | commit 2-2 | commit 2-3\n2023-02-20 16:29:39+03:00 | Nikolay Mikhaylichenko\n[\'branch-3\', \'main\']\nAdded lines:        8\nDeleted lines:      0\nChanged files:      1\n```\n\n### codechanges\n\nShow total added and deleted lines count for current ref.\n\n```bash\n# --- show command help\n$ smyg codechanges --help\n\n# --- changes for current ref\n$ smyg codechanges\nRatio (%):         11\nAdded lines:      107\nDeleted lines:     12\n\n# --- code changes between commits\n$ smyg codechanges --from-commit 3a085582e8df7c1d622c412eb54eb5ee96e82c48  \\\n  --to-commit bbead3ba03de57909ee622fac4b973de180f5e5f\nRatio (%):         42\nAdded lines:       26\nDeleted lines:     11\n```\n\n### codechurn\n\nShow count of added lines and count of deleted lines from those that were added. Displays the level of work that was not released.\n\n> Ideally, it should coincide with the codechanges for whole project history. For complex branching it may slightly different.\n\n```bash\n# --- show command help\n$ smyg codechurn --help\n\n# --- churn for current ref (equal with codechanges, it is expected)\nsmyg codechurn\nRatio (%):         11\nAdded lines:      107\nDeleted lines:     12\n\n# --- churn from specified commit (not equal with codechanges, it is expected)\n$ smyg codechurn --from-commit 44c6334f38fe796f38c39e1d11a8d19925684926 \\\n    --to-commit fccfa86625353409018ccd467f0050b57152656e\nRatio (%):         23\nAdded lines:       26\nDeleted lines:      6\n```\n\n## Prometheus Metrics\n\nYou can push metrics to Prometheus PushGateway. \nRecommended to use Aggregation Gateway like https://github.com/zapier/prom-aggregation-gateway.\n\n`commit` and `branch_commits` is a *Counter* type, `codechanges` and `codechurn` - *Gauge*. \n\nYou must specify environment variables `PUSHGATEWAY_URL` and `PROJECT_NAME` and command option `--push-metrics` to send metrics. For basic auth access specify `PUSHGATEWAY_USERNAME` and `PUSHGATEWAY_PASSWORD`.\n\n```bash\n$ git checkout main\n\n$ docker run --rm -p 8080:80 ghcr.io/zapier/prom-aggregation-gateway:v0.7.0\n\n$ PUSHGATEWAY_URL=localhost:8080 PROJECT_NAME=example_project \\\n    smyg commit --push-metrics\ncdcb7c6ac0b92b1e16cfdd6f1e6abf0ed8f73e48 | update README\n2023-02-20 16:42:41+03:00 | Nikolay Mikhaylichenko\n[\'main\']\nAdded lines:       26\nDeleted lines:      0\nChanged files:      1\n\n$ curl localhost:8080/metrics\n# HELP commit_added_lines_total Number of added lines in commit\n# TYPE commit_added_lines_total counter\ncommit_added_lines_total{author_email="nn.mikh@yandex.ru",author_name="Nikolay Mikhaylichenko",job="showmeyourgit",project_name="example_project"} 26\n# HELP commit_changed_files_total Number of changed files in commit\n# TYPE commit_changed_files_total counter\ncommit_changed_files_total{author_email="nn.mikh@yandex.ru",author_name="Nikolay Mikhaylichenko",job="showmeyourgit",project_name="example_project"} 1\n# HELP commit_deleted_lines_total Number of added deleted in commit\n# TYPE commit_deleted_lines_total counter\ncommit_deleted_lines_total{author_email="nn.mikh@yandex.ru",author_name="Nikolay Mikhaylichenko",job="showmeyourgit",project_name="example_project"} 0\n# HELP commits_total Number of commits\n# TYPE commits_total counter\ncommits_total{author_email="nn.mikh@yandex.ru",author_name="Nikolay Mikhaylichenko",job="showmeyourgit",project_name="example_project"} 1\n```\n\n## Develop & Test\n\n```bash\n$ git clone git@github.com:nmix/showmeyourgit.git\n$ cd showmeyourgit\n$ poetry install\n$ poetry shell\n$ pytest\n```\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first\nto discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Nikolay Mikhaylichenko',
    'author_email': 'nn.mikh@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nmix/showmeyourgit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



'''Package bindings'''

from __future__ import annotations

import sys
import git
import pydriller

from smyg import vcs


class BindingError(Exception):
    '''Default exception for binding module'''


def branch_commits(
        path: str,
        sha: str | None = None,
        branch: str | None = None
        ) -> list[vcs.Commit]:
    '''Find bundle of commits for a branch

    Args:
        path: absolute path to repo
        sha: starting commit SHA
            if sha not null then returns all commits from this sha
            if sha null then return all commits only for a branch
        branch: specified branch
    '''

    if sha:
        repo = pydriller.Repository(path,
                                    from_commit=sha,
                                    only_in_branch=branch,
                                    order='reverse')
        commits = list(repo.traverse_commits())[:-1]
    else:
        repo = pydriller.Repository(path,
                                    only_in_branch=branch,
                                    order='reverse')
        commits = []
        for commit in repo.traverse_commits():
            if len(commit.branches) == 1:
                commits.append(commit)
            else:
                break
    # ---
    return [_create_vcs_commit(commit) for commit in commits]


def find_commit(
        path: str,
        sha: str | None = None,
        branch: str | None = None):
    '''Find commit in history'''
    if sha:
        repo = pydriller.Repository(path,
                                    single=sha,
                                    only_in_branch=branch,
                                    order='reverse')
    else:
        repo = pydriller.Repository(path,
                                    only_in_branch=branch,
                                    order='reverse')
    try:
        commit = next(repo.traverse_commits())
    except git.GitCommandError as e:
        raise BindingError from e
    return _create_vcs_commit(commit)


def modified_files(path: str, **kargs):
    '''Get list of modified files from repo

    Attributes:
        path: repo path
    '''
    files = []
    # --- TODO: try except in for next() iteration instead separate call
    try:
        next(pydriller.Repository(path, **kargs).traverse_commits())
    except git.GitCommandError as e:
        raise BindingError from e
    except StopIteration:
        pass
    # ---
    for commit in pydriller.Repository(path, **kargs).traverse_commits():
        try:
            commit_files = commit.modified_files
        except ValueError as e:
            print(f'Error occurred while processing the commit '
                  f'{commit.sha}: {e}', file=sys.stderr)
            continue
        for file in commit_files:
            files.append(
                    vcs.ModifiedFile(
                        old_path=file.old_path,
                        new_path=file.new_path,
                        added_lines=file.diff_parsed.get('added'),
                        deleted_lines=file.diff_parsed.get('deleted')))
    return files


def _create_vcs_commit(commit: pydriller.Commit) -> vcs.Commit:
    return vcs.Commit(
            sha=commit.hash,
            msg=commit.msg,
            author_name=commit.author.name,
            author_email=commit.author.email,
            author_date=commit.author_date,
            committer_name=commit.committer.name,
            committer_email=commit.committer.email,
            committer_date=commit.committer_date,
            project_name=commit.project_name,
            added=commit.insertions,
            deleted=commit.deletions,
            changed_files=commit.files,
            branches=commit.branches,
            )
